C!!! NASA Goddard Space Flight Center Open Source: Jan McGarry (Jan.McGarry@nasa.gov)
C =======================================================================
C STAR ANALYSIS ROUTINES:  Jan McGarry/NASA/GSFC/920.1
C       1. RECURSIVE    Kalman filter.
C       2. BATCH        Least squares batch process: calling routine.
C          UPDATE       Perform the least squares solution.
C          CHKERR       Check the error in the least squares fit.
C       3. SUMMARY      Summary computations and file write.
C       4. COVINI       Initialize the covariance matrix (and coeffs).
C       5. COEWRT       Write out the coefficients to file.
C       6. COVWRT       Write out the covariance matrix to file.
C       7. DEF_READ     Read the [STAR_DEFAULT.DAT] file.
C       8. MMODEL       Compute the mount model correction.
C =======================================================================
C =======================================================================
        SUBROUTINE RECURSIVE(I,SDF,IRETF)
C jlfm/GSFC/920.1: 4/5/93...6/1/93...6/18/93...structure added 6/23/93
C       correction of SIGR (make it squared): jlfm 1/2/94.
C -----------------------------------------------------------------------
C Mount Model recursive update routine: KALMAN FILTER technique.
C                (see Gelb, pp.102-110)
C       Inputs:  A1:     Encoder angle #1 (AZ or X): raw + offset
C                A2:     Encoder angle #2 (EL or Y): raw + offset
C                D1:     Angle #1 bias entered to center this star
C                D2:     Angle #2 bias entered to center this star
C                SIGR:   Standard deviation of measurement noise.
C                NMMC:   Number of mount model coeffs.
C                X:      Current mount model coefficients
C                P:      Current covariance matrix
C
C       Outputs: X:      Updated mount model coefficients
C                P:      Updated covariance matrix
C                IRETF:  Return flag, = 0 update occurred OK.
C                                     = 1 update did not occur. 
C
C         W  = H*P(-)*HT + R
C         K  = P(-) * HT * W**-1
C       x(+) = x(-) + K*ZMHX   where ZMHX = [z - H*x]
C       P(+) = P(-) - K*H*P(-)
C
C       Matrix dimensions:  
C                    [x] NMMC x 1
C                    [P] NMMC x NMMC
C                    [H] 2 x NMMC
C                    [K] NMMC x 2
C                    [W] 2 x 2
C
C       Note:  (1) "R" is assumed to be the constant "SIGR*SIGR" times
C                       the identity matrix.
C              (2) "z" is the measurement (negative of the bias angles).
C -----------------------------------------------------------------------
        IMPLICIT REAL*8 (A-H,O-Z)
        INCLUDE 'STARFILE.INC'
        TYPE (STAR_DATA_FILE_STRUCT) SDF
        PARAMETER (RAD=57.295779513D0, MMAX=30)
        REAL*8 ZMHX(2),W(2,2),WINV(2,2)
        REAL*8 K(2*MMAX),H(2*MMAX),TEMP1(2*MMAX),TEMP2(2*MMAX)
C
C???        I=SDF%NSTARS ???
C       /// Produce the matrix "H" ///
C       * station specific subroutine *
        CALL HMODEL(SDF%SD(I)%A1,SDF%SD(I)%A2,H)
C
C       /// Computations for matrix "K" follow ///
C           TEMP1 is 2xM
        CALL MATMUL(H,SDF%P,TEMP1,2,NMMC,NMMC,1,0)
        CALL MATMUL(TEMP1,H,W,2,NMMC,2,3,0)
        W(1,1)=W(1,1)+SDF%SIGR**2
        W(2,2)=W(2,2)+SDF%SIGR**2
C          Compute the inverse of this 2x2 symmetric matrix.
        DET=W(1,1)*W(2,2)-W(1,2)*W(1,2)
        IF(DABS(DET).LT.1.D-16) GOTO 500
        WINV(1,1)= W(2,2)/DET
        WINV(1,2)=-W(1,2)/DET
        WINV(2,2)= W(1,1)/DET
        WINV(2,1)= WINV(1,2)
C          TEMP2 is Mx2
        CALL MATMUL(H,WINV,TEMP2,NMMC,2,2,2,0)
        CALL MATMUL(SDF%P,TEMP2,K,NMMC,NMMC,2,1,0)
C
C       /// Produce the measurement array minus the current state ///
        ZMHX(1)=SDF%SD(I)%D1
        ZMHX(2)=SDF%SD(I)%D2
C
C       /// Update the mount model coefficients ///
        CALL MATMUL(K,ZMHX,SDF%X,NMMC,2,1,1,1)
C
C       /// Update the covariance matrix ///
        CALL MATMUL(K,TEMP1,SDF%P,NMMC,2,NMMC,1,-1)
C
C       /// Update OK return flag ///
        IRETF=0
        RETURN
C
C       /// Update did not occur ///
500     IRETF=1
        RETURN
        END

C =======================================================================

        SUBROUTINE BATCH(SDF,MNTTYP,IRETF)
C jlfm/GSFC/920.1: 4/5/93...5/31/93...6/18/93...structure added on 6/23/93
C       ...SIGR correction 8/11/93 (jlfm)
C ---------------------------------------------------------------
C Mount Model BATCH update routine: LEAST-SQ with apriori knowledge
C                (see Maybeck, pp.117-121).
C       Inputs:  STARNUM:  Star number.
C                A1:     Encoder angle #1 (AZ or X): raw + offset.
C                A2:     Encoder angle #2 (EL or Y): raw + offset.
C                B1:     Total angle #1 bias applied to center this star.
C                        ie sum of input bias plus mount model correction.
C                B2:     Total angle #2 bias applied to center this star.
C                        ie sum of input bias plus mount model correction.
C                NSTARS: Number of measurements.
C                SIGR:   Standard deviation of measurement noise.
C                ITER:   Maximum number of iterations (passes) thru data.
C                SIGSCL: Multiplier for STD, to determine data cutoff.
C                NMMC:   Number of mount model coeffs.
C                X:      Current mount model coefficients: x(-).
C                P:      Current covariance matrix: P(-).
C                MNTTYP: Mount type: 0==>AZ/EL (48",MOBLAS,TLRS3/4)
C                                    1==>AZ/Zenith (LURE)
C                                    3==>X-Y (MLRS)
C
C       Outputs: STARNUM:   Stars rejected will have negative numbers
C                           returned in the star list.
C                X1:     Updated mount model coefficients: x(+).
C                P1:     Updated covariance matrix: P(+).
C
C         W  = [ sigr**2*P(-)**-1 + sum(HT*H) ]
C       P(+) = W**-1/sigr**2
C       x(+) = p + q
C          where p = [P(+)*P(-)**-1]*x(-)
C                q = (W**-1)*sum(HT*z)  and "z" is the measurement.
C                and where sum(.) is meant to imply sum over all
C                        NSTARS measurements.
C
C       Matrix dimensions:  
C                    [x] NMMC x 1
C                    [P] NMMC x NMMC
C                    [H] 2 x NMMC
C                    [W] NMMC x NMMC
C                    [TEMP] NMMC x NMMC
C
C       Note:  "R" is assumed to be the constant "SIGR*SIGR" times the
C              identity matrix.
C ---------------------------------------------------------------
        IMPLICIT REAL*8 (A-H,O-Z)
        INCLUDE 'STARFILE.INC'
        TYPE (STAR_DATA_FILE_STRUCT) SDF
        PARAMETER (MMAX=30, NMAX=200)
        REAL*8 X1(MMAX), P1(MMAX*MMAX)
C
C          Eliminate all multiple star occurences...
C          ensure we take only the last one.
        DO KK=1,SDF%NSTARS-1
           IF(SDF%SD(KK)%STARNUM.GT.0.AND.
     .       (SDF%SD(KK)%STARNUM.EQ.SDF%SD(KK+1)%STARNUM)) THEN
                  SDF%SD(KK)%STARNUM=-ABS(SDF%SD(KK)%STARNUM)
           ENDIF
        END DO
        NOLD=0
        SDF%NREJ=0
        KREJF=0
        IRETF=0
        DO KK=1,SDF%ITER+1
C               compute new state and covariance: X1,P1
           CALL UPDATE(SDF,X1,P1,IRETF)
           IF(IRETF.NE.0) GOTO 510
C               indicate last iteration
           IF(KK.EQ.SDF%ITER+1) KREJF=1
C               compute error in model fit, and reject stars
C                       if not last iteration
           CALL CHKERR(SDF,X1,P1,MNTTYP,KREJF)
           IF(SDF%NOK.LT.2) THEN
C               * number of stars is too small *
              print *,'Error - number of stars is too small'
              IRETF=2
              GOTO 510
           ENDIF
C               if no stars were rejected, then DONE!
           IF(SDF%NREJ.EQ.NOLD) GOTO 500
           NOLD=SDF%NREJ
        END DO
C          * move over state vector and covariance for return *
500     DO I=1,NMMC
           SDF%X(I)=X1(I)
        END DO
        CALL MATCOPY(P1,SDF%P,NMMC,NMMC)
510     RETURN
        END

        SUBROUTINE UPDATE(SDF,X1,P1,IFLAG)
C jlfm/GSFC/920.1...6/1/93...structure added on 6/23/93
C       ...SIGR correction 8/11/93 (jlfm)
C ---------------------------------------------------------------
C   BATCH PROCESS UPDATE (using apriori knowledge of covariance matrix).
C
C       IFLAG=0 on RETURN if UPDATE was successful.
C ---------------------------------------------------------------
        IMPLICIT REAL*8 (A-H,O-Z)
        INCLUDE 'STARFILE.INC'
        TYPE (STAR_DATA_FILE_STRUCT) SDF
        PARAMETER (RAD = 57.295779513D0, MMAX=30)
        DIMENSION X1(NMMC),P1(NMMC,NMMC)
        DIMENSION Z(2)
        DIMENSION PINV(MMAX*MMAX),TEMP(MMAX*MMAX),W(MMAX*MMAX)
        DIMENSION H(2*MMAX),XP(MMAX),XQ(MMAX)
C
        DO I=1,NMMC
           XQ(I)=0.D0
        END DO
C
C       /// Computations for matrix "W" follow ///
C       First invert the incoming covariance matrix: P(-)**-1
        CALL MATINV(SDF%P,NMMC,PINV,IFLAG)
        IF(IFLAG.NE.0) THEN
	     print *,'Error in UPDATE: matrix inversion #1'
	     GOTO 500
        ENDIF
C       /// W = sigr**2/P(-)
        DO I=1,NMMC*NMMC
           W(I)=SDF%SIGR**2*PINV(I)
        END DO
C
C       /// Produce the running sum matrix "HT*H" over N measurements ///
        DO I=1,SDF%NSTARS
           IF(SDF%SD(I)%STARNUM.GT.0) THEN
C               * station specific subroutine *
              CALL HMODEL(SDF%SD(I)%A1,SDF%SD(I)%A2,H)
C               * running sum: W = W + HT*H *
              CALL MATMUL(H,H,W,NMMC,2,NMMC,2,1)
C               * measurement vector "z" *
              Z(1)=SDF%SD(I)%B1
              Z(2)=SDF%SD(I)%B2
C               * running sum: XQ = XQ + HT*Z *
              CALL MATMUL(H,Z,XQ,NMMC,2,1,2,1)
           ENDIF
        END DO
C
C       /// Compute P(+) = sigr**2/W///
        CALL MATINV(W,NMMC,P1,IFLAG)
        IF(IFLAG.NE.0) THEN
           print *,'Error in UPDATE: matrix inversion #2'
	     GOTO 500
        ENDIF
        DO I=1,NMMC
           DO J=1,NMMC
              P1(I,J)=P1(I,J)*SDF%SIGR**2
           END DO
        END DO
C
C       /// Compute "x(+)" ///
C       Compute "p" first...
        CALL MATMUL(SDF%P,W,TEMP,NMMC,NMMC,NMMC,1,0)
        DO I=1,NMMC*NMMC
           TEMP(I)=TEMP(I)/SDF%SIGR**2
        END DO
        DO I=1,NMMC
           XP(I)=SDF%X(I)
        END DO
        CALL SOLVE(TEMP,XP,NMMC,IFLAG)
        IF(IFLAG.NE.0) THEN
	     print *,'Error return from SOLVE #1'
	     GOTO 500
        ENDIF
C       Compute "q" next...
        CALL SOLVE(W,XQ,NMMC,IFLAG)
        IF(IFLAG.NE.0) THEN
	     print *,'Error return from SOLVE #2'
	     GOTO 500
	  ENDIF
C       Add to get "x(+)"
        DO I=1,NMMC
           X1(I)=XP(I) + XQ(I)
        END DO
        IFLAG=0
500     RETURN
        END


        SUBROUTINE CHKERR(SDF,X,P,MNTTYP,KLAST)
C jlfm/GSFC/920.1: 6/1/93...6/18/93...structure added on 6/23/93
C -------------------------------------------------------------------
C    Compute model fit error and reject the star with the largest
C       error above SIGSCL*RMSTOT.  If KLAST is nonzero, then just
C       compute the fit error, but don't reject any stars.
C
C       Note: stars are rejected by changing the star number STARNUM
C             to negative.
C -------------------------------------------------------------------
        IMPLICIT REAL*8 (A-H,O-Z)
        INCLUDE 'STARFILE.INC'
        TYPE (STAR_DATA_FILE_STRUCT) SDF
        PARAMETER (RAD = 57.295779513D0, NMAX=200, MMAX=30)
        DIMENSION X(NMMC),P(NMMC,NMMC)
        DIMENSION RMSERR(NMAX)
        DIMENSION Z(2),H(2*MMAX)
C
        SUM=0.D0
        SUMSQ=0.D0
        SDF%NOK=0
C
C       /// Compute fit error ///
        DO I=1,SDF%NSTARS
C               only work with non-rejected stars
           IF(SDF%SD(I)%STARNUM.GT.0) THEN
              Z(1)=SDF%SD(I)%B1
              Z(2)=SDF%SD(I)%B2
C
              IF(MNTTYP.EQ.0) THEN
                XSKY=DCOS(SDF%SD(I)%A2/RAD)
                YSKY=1.D0
              ELSEIF(MNTTYP.EQ.1) THEN
                XSKY=DSIN(SDF%SD(I)%A2/RAD)
                YSKY=1.D0
              ELSEIF(MNTTYP.EQ.2) THEN
                XSKY=1.D0
                YSKY=DCOS(SDF%SD(I)%A1/RAD)
              ELSEIF(MNTTYP.EQ.3) THEN
                XSKY=1.D0
                YSKY=DSIN(SDF%SD(I)%A1/RAD)
              ENDIF
C               compute matrix "H": station specific routine.
              CALL HMODEL(SDF%SD(I)%A1,SDF%SD(I)%A2,H)
C                  z = z - H*x
              CALL MATMUL(H,X,Z,2,NMMC,1,1,-1)
              RMSERR(I) = Z(1)*Z(1)*XSKY*XSKY + Z(2)*Z(2)*YSKY*YSKY
C               running sum for RMS and mean calculations
              SUMSQ = SUMSQ + RMSERR(I)
              RMSERR(I) = DSQRT(RMSERR(I))
              SUM = SUM + Z(1)*XSKY + Z(2)*YSKY
              SDF%NOK=SDF%NOK+1
           ENDIF
        END DO
        IF(SDF%NOK.LT.2) GOTO 500
        SDF%SMEAN=SUM/SDF%NOK
        SDF%RMSTOT=DSQRT(SUMSQ/(SDF%NOK-1))
C/STD.DEV/     SDF%RMSTOT=DSQRT((SUMSQ-SDF%NOK*SMEAN)/(SDF%NOK-1))
        RMSMAX=0.D0
        LREJ=0
        IF(KLAST.EQ.0) THEN
C
C       /// Find the star with the largest error outside our limit ///
           DO I=1,SDF%NSTARS
C               Only look at non-rejected stars with
C               errors greater than SCL*RMSTOT.
              IF(SDF%SD(I)%STARNUM.GT.0.AND.
     .              RMSERR(I).GT.SDF%SIGSCL*SDF%RMSTOT) THEN
                 IF(RMSERR(I).GT.RMSMAX) THEN
                    RMSMAX=RMSERR(I)
                    LREJ=I
                 ENDIF
              ENDIF
           END DO
C               If any star fell outside our limit, then
C               pick up the biggest and reject it.
           IF(LREJ.NE.0) THEN
              SDF%NREJ=SDF%NREJ+1
              SDF%SD(LREJ)%STARNUM=-ABS(SDF%SD(LREJ)%STARNUM)
           ENDIF
        ENDIF
500     RETURN
        END
 
C =======================================================================

        SUBROUTINE SUMMARY(SDF,X0,P0,MNTTYP,CURDATE)
C jlfm/GSFC/920.1...6/1/93...6/18/93...structure added 6/23/93
C       Cosmetic corrections: 10/10/95
C--------------------------------------------------------------------
C ANALYSIS OF STAR CALIBRATION DATA SET AND SUMMARY OUTPUT TO FILE
C       * CALLED AT THE END OF A STAR CALIBRATION *
C
C       SYSNAME         System name (30 character string)
C       NSTARS          Number of stars in the star calibration
C       STARNUM         Star number
C       CATALOG_NUM     Corresponding catalog (FK5) numbers
C       A1              Array of angles (AZ or X)
C       A2              Array of angles (EL or Y)
C       B1              Array of total angle 1 biases (AZ or X)
C                       ie sum of input bias plus mount model correction
C       B2              Array of total angle 2 biases (EL or Y)
C                       ie sum of input bias plus mount model correction
C       D1              Array of entered angle 1 bias (input bias)
C       D2              Array of entered angle 2 bias (input bias)
C       NMMC            Number of mount model coefficients
C       X0              Initial mount model coefficients
C       P0              Initial covariance matrix
C       X               Mount model coefficients after analysis
C       P               Covariance matrix after analysis
C       MNTTYP          Mount type (0=AZ/EL,1=AZ/Zenith,3=X-Y)
C       FITTYPE         Type of fit (BATCH=0, KF=1)
C       SIGR            Estimated measurement error
C       ITER            Number of iterations of star rejections
C       SIGSCL          Scale factor (mult.of STD) for star rejection
C
C       FILENAME        Filename for analysis file
C       CATALOG         Catalog name (ASCII string)
C       DATAFILE        Filename (includes date) of star cal data
C       CURDATE         Current date of analysis (ASCII string: mm-dd-yyyy)
C       TEMP            Temperature at star calibration time (deg.C)
C       BARP            Barometric pressure at starcal time (millibars)
C       HUMD            Humidity at starcal time (%)
C
C--------------------------------------------------------------------
        IMPLICIT REAL*8 (A-H,O-Z)
        INCLUDE 'STARFILE.INC'
        TYPE (STAR_DATA_FILE_STRUCT) SDF
C               radians to degrees
        PARAMETER (RAD = 57.295779513D0)
C               degrees to arcseconds
        PARAMETER (F = 3600.D0)
C--------------------------------------------------------------------
        CHARACTER*30 FILENAME
        CHARACTER*10 CURDATE
        REAL*8 X0(NMMC),P0(NMMC,NMMC)
        INTEGER I,NSUM
C
        FILENAME(1:5)  = 'anal_'
        FILENAME(6:27) = SDF%FILENAME(6:27)
        FILENAME(28:30)= '   '
        OPEN(UNIT=10,FILE=FILENAME,STATUS='UNKNOWN')
 1010   FORMAT(1X,A12,A10)
 1020   FORMAT('  Weather (T/B/H): ',F7.3,1X,F8.3,1X,F5.1,/)
C
C               * HEADER *
        WRITE(10,'(X,A)') '* STAR CALIBRATION ANALYSIS SUMMARY FILE *'
        WRITE(10,'(1X,A)') SDF%SYSNAME
        WRITE(10,1010) 'CATALOG:    ',SDF%CATALOG(1:10)
        IF(SDF%FITTYPE.EQ.0) THEN
           WRITE(10,1010) 'ANALYSIS:   ','GLOBAL    '
        ELSE
           WRITE(10,1010) 'ANALYSIS:   ','KALMAN    '
        ENDIF
        WRITE(10,'(A12,A30)') ' Data file:  ',SDF%FILENAME
        WRITE(10,1010) 'Analysis on ',CURDATE
        WRITE(10,'(A20,F9.6)') '  Est.meas.error:   ',SDF%SIGR
        WRITE(10,'(A20,I3)')   '  # iterations:     ',SDF%ITER
        WRITE(10,'(A20,F6.3)') '  Rej.scale factor: ',SDF%SIGSCL
        WRITE(10,1020) SDF%TEMP,SDF%BARP,SDF%HUMD
        WRITE(10,1090)
 1090   FORMAT(/,' * INITIAL COEFFICIENTS AND STANDARD DEVIATIONS *')
        DO I=1,NMMC
           WRITE(10,1050) X0(I),DSQRT(P0(I,I))
        END DO
        WRITE(10,*) ' '
C
C               * DATA *
        SUMDD1=0.D0
        SUMDD2=0.D0
        SUMRES=0.D0
        SSDD1 =0.D0
        SSDD2 =0.D0
        SSRES =0.D0
        NSUM  =0
        WRITE(10,1080)
 1080   FORMAT(' STAR#  CAT# SKY# HH:MM:SS.SS    ANG1  TOTBIAS1 ',
     .       'INP.BIAS1   ANG2  TOTBIAS2 INP.BIAS2  DELTA1    ',
     .       'DELTA2    SKYRES')
        DO I=1,SDF%NSTARS
           IF(MNTTYP.EQ.0) THEN
                XSKY=DCOS(SDF%SD(I)%A2/RAD)
                YSKY=1.D0
           ELSEIF(MNTTYP.EQ.1) THEN
                XSKY=DSIN(SDF%SD(I)%A2/RAD)
                YSKY=1.D0
           ELSEIF(MNTTYP.EQ.2) THEN
                XSKY=1.D0
                YSKY=DCOS(SDF%SD(I)%A1/RAD)
           ELSEIF(MNTTYP.EQ.3) THEN
                XSKY=1.D0
                YSKY=DSIN(SDF%SD(I)%A1/RAD)
           ENDIF
           CALL MMODEL(SDF%X,NMMC,SDF%SD(I)%A1,SDF%SD(I)%A2,DD1,DD2)
           DD1 = (SDF%SD(I)%B1 - DD1) * XSKY
           DD2 = (SDF%SD(I)%B2 - DD2) * YSKY
           RES = DSQRT(DD1*DD1 + DD2*DD2)
           IF(SDF%SD(I)%STARNUM.GT.0) THEN
                NSUM = NSUM + 1
                SUMDD1=SUMDD1+DD1
                SUMDD2=SUMDD2+DD2
                SUMRES=SUMRES+RES
                SSDD1 = SSDD1 + DD1*DD1
                SSDD2 = SSDD2 + DD2*DD2
                SSRES = SSRES + RES*RES
           ENDIF
           WRITE(10,1030) SDF%SD(I)%STARNUM,
     .                    SDF%SD(I)%CATALOG_NUM,
     .                    SDF%SD(I)%SKY_SECTION,
     .                    SDF%SD(I)%TOD%HRS,
     .                    SDF%SD(I)%TOD%MINS,
     .                    SDF%SD(I)%TOD%SECS,
     .                    SDF%SD(I)%A1,SDF%SD(I)%B1,SDF%SD(I)%D1,
     .                    SDF%SD(I)%A2,SDF%SD(I)%B2,SDF%SD(I)%D2,
     .                    DD1,DD2,RES
 1030      FORMAT(1X,I4,1X,I5,1X,I4,1X,I2.2,':',I2.2,':',F5.2,
     .                    6(1X,F8.4),3(1X,F9.6))
        END DO
C
C               * SUMMARY *
        XMUDD1=SUMDD1/NSUM
        XMUDD2=SUMDD2/NSUM
        XMURES=SUMRES/NSUM
        DGFREE=(2*NSUM)**2/(3*NSUM+NMMC)
        RMSDD1=DSQRT(SSDD1/DGFREE)
        RMSDD2=DSQRT(SSDD2/DGFREE)
        RMSRES=DSQRT(SSRES/DGFREE)
        SDF%SMEAN =XMURES
        SDF%RMSTOT=RMSRES*F
        SDF%NOK   =NSUM
        SDF%NREJ  =SDF%NSTARS-NSUM
        WRITE(10,1091)
1091    FORMAT(/,' * STAR CALIBRATION SUMMARY *')
        WRITE(10,1093) SDF%NSTARS,SDF%NREJ,NSUM
1093    FORMAT(' Total number of stars:       ',I3,/,
     .         ' Number of stars rejected:    ',I3,/,
     .         ' Actual number of stars used: ',I3,/)
        WRITE(10,1094)
1094    FORMAT('          MEAN(deg) MEAN(asec) RMS(deg)  RMS(asec)')
        WRITE(10,1040) 'ANG #1: ',XMUDD1,XMUDD1*F,RMSDD1,RMSDD1*F
        WRITE(10,1040) 'ANG #2: ',XMUDD2,XMUDD2*F,RMSDD2,RMSDD2*F
        WRITE(10,1040) 'SKYANG: ',XMURES,XMURES*F,RMSRES,RMSRES*F
1040    FORMAT(1X,A8,F9.6,1X,F9.3,1X,F9.6,1X,F9.3)
C
C               * COEFFICIENTS AND STANDARD DEVIATIONS *
        WRITE(10,1092)
 1092   FORMAT(/,' * COEFFICIENT SUMMARY *',/,
     .                  ' COEFFICIENTS      STD.DEV')
        DO I=1,NMMC
	     if(SDF%P(I,I).lt.0) SDF%P(I,I)=0.0
           WRITE(10,1050) SDF%X(I),DSQRT(SDF%P(I,I))
 1050      FORMAT(1X,F12.8,1X,F12.8)
        END DO
C
        CLOSE(UNIT=10)
        RETURN
        END

        SUBROUTINE SECTOHMS(TIME,IH,IM,IS)
C ---------------------------------------------------------
C  CONVERT SECONDS TO HOURS:MINUTES:SECONDS
C ---------------------------------------------------------
        REAL*8 TIME
        INTEGER IH,IM,IS
        IH=INT(TIME/3600.D0)
        IM=INT((TIME-IH*3600.D0)/60.D0)
        IS=INT(TIME-IH*3600.D0-IM*60.D0 + 0.5)
        RETURN
        END

C ==========================================================================

        SUBROUTINE COVINI(IOPT,SDF,PINIT,MASK)
C jlfm/GSFC/920.1...6/1/93...structure added on 6/23/93
C       IOPT=2 added on 8/9/93(jlfm).
C       Scaline the P-Matrix using a MASK array: jlfm...1/17/94.
C--------------------------------------------------------------------
C INITIALIZE THE COVARIANCE MATRIX
C       CALLED AT START OF A STAR CALIBRATION
C
C       IOPT = 0 ==> set coefficients and covariance to the default
C                       initial (no apriori) state.
C            = 1 ==> use existing coefficients and scale covariance.
C            = 2 ==> use existing coefficients and set covariances to
C                       the initial state.
C       NMMC         Number of mount model coefficients.
C       X            Mount model coeff array.
C       P            Covariance matrix
C       PINIT        Initial covariance matrix (when starting the starcal
C                       from scratch).
C       PSCALE       Multiplier for an existing covariance matrix.
C       MASK         Covariance matrix mask array; for IOPT=1 scaling.
C                       (Zero ==> scale normally; nonzero ==> init row/colm).
C--------------------------------------------------------------------
        INCLUDE 'STARFILE.INC'
        TYPE (STAR_DATA_FILE_STRUCT) SDF
        REAL*8 PINIT(NMMC)
        INTEGER MASK(NMMC)
        INTEGER IOPT,I,J
C
        IF(IOPT.EQ.0) THEN
           DO I=1,NMMC
              DO J=1,NMMC
                 SDF%P(I,J)=0.D0
              END DO
              SDF%X(I)=0.D0
              SDF%P(I,I)=PINIT(I)
           ENDDO           
        ELSEIF(IOPT.EQ.1) THEN
           DO I=1,NMMC
              DO J=1,NMMC
                 SDF%P(I,J)=SDF%P(I,J)*SDF%PSCALE
              END DO
           END DO
C   Added on 1/17/94 (mask off scaling parts of covariance matrix)
           DO I=1,NMMC
              IF(MASK(I).NE.0) THEN
                 DO J=1,NMMC
                    IF(J.EQ.I) THEN
                        SDF%P(I,J)=PINIT(I)
                    ELSE
                        SDF%P(I,J)=0.D0
                        SDF%P(J,I)=0.D0
                    ENDIF
                 ENDDO
              ENDIF
           END DO
        ELSEIF(IOPT.EQ.2) THEN
           DO I=1,NMMC
              DO J=1,NMMC
                 SDF%P(I,J)=0.D0
              END DO
              SDF%P(I,I)=PINIT(I)
           END DO
        ENDIF
        RETURN
        END

C ==========================================================================

        SUBROUTINE COEWRT(NMMC,X,RMS,FILENAME)
C jlfm/GSFC/920.1...6/1/93
C--------------------------------------------------------------------
C WRITE MOUNT MODEL COEFFICIENTS TO FILE
C--------------------------------------------------------------------
        DOUBLE PRECISION RMS
        CHARACTER*30 FILENAME
        REAL*8 X(NMMC)
        INTEGER I
        OPEN(UNIT=10,FILE=FILENAME,STATUS='UNKNOWN')
        WRITE(10,'(I3,3X,F9.6)') NMMC,RMS
        DO I=1,NMMC
           WRITE(10,'(F12.8)') X(I)
        ENDDO
        CLOSE(UNIT=10)
        RETURN
        END

C ==========================================================================

        SUBROUTINE COVWRT(NMMC,P,FILENAME)
C jlfm/GSFC/920.1...6/1/93
C--------------------------------------------------------------------
C WRITE MOUNT MODEL COVARIANCE MATRIX TO FILE
C       * ONLY THE UPPER TRIANGULAR PART *
C       NOTE THAT THIS IS A SYMMETRIC MATRIX!
C--------------------------------------------------------------------
        CHARACTER*30 FILENAME
        REAL*8 P(NMMC,NMMC)
        INTEGER I
        OPEN(UNIT=10,FILE=FILENAME,STATUS='UNKNOWN')
        WRITE(10,'(I3)') NMMC
        DO I=1,NMMC
           WRITE(10,1010) (P(J,I),J=I,NMMC)
 1010      FORMAT(28(1X,F12.8))
        ENDDO
        CLOSE(UNIT=10)
        RETURN
        END

C ==========================================================================

        SUBROUTINE DEF_READ(NMMC,SYSNAME,MAXSTR,UPD_TYPE,BSMAGL,BSMAGU,
     .             GSMAGL,GSMAGU,IASTAR,SIGR,ITER,SIGSCL,PSCALE,
     .             IOPT,PINIT,MASK)
C jlfm/GSFC/920.1...6/1/93...6/23/93...1/17/94(MASK)...10/5/95(NSARC)
C--------------------------------------------------------------------
C READ THE DEFAULT STAR FILE [STAR_DEFAULT.DAT].
C
C       NMMC:      Number of mount model coefficients.
C       SYSNAME:   System name (30 character string).
C       MAXSTR:    Max. number of stars allowed in star_table (up to 200).
C       UPD_TYPE:  Fit type (BRIGHT, GRID, AUTO...)
C       BSMAGL:    Bright stars lower magnitude limit.
C       BSMAGU:    Bright stars upper magnitude limit.
C       GSMAGL:    Grid stars lower magnitude limit.
C       GSMAGU:    Grid stars upper magnitude limit.
C       IASTAR:    Alignment star number (from FK5 catalog).
C       NSARC:     Number of stars in satellite arc.
C       SIGR:      Estimate of measurement error (degrees).
C       ITER:      Number of star rejection iterations.
C       SIGSCL:    Multiplier for STD in star error to reject stars.
C       PSCALE:    Scale factor to roundup an existing P at start.
C       IOPT:      Initial IOPT value for COVINI.
C       PINIT:     Initial covariance matrix (diag only) for starcal
C                       performed from scratch.
C       MASK:      Mask for IOPT=1 covariance setup, used in COVINI.
C       
C--------------------------------------------------------------------
        IMPLICIT REAL*8 (A-Z)
        CHARACTER*30 SYSNAME
        CHARACTER*8 UPD_TYPE
	CHARACTER   DUM
        INTEGER NMMC,MAXSTR,ITER,I,IASTAR,NSARC,IOPT
        DIMENSION PINIT(NMMC),PMASK(30)
        INTEGER MASK(NMMC)
        DO I=1,30
           SYSNAME(I:I)=CHAR(0)
        END DO
        OPEN(UNIT=10,FILE='star_default.dat',STATUS='OLD')
        READ(10,'(A)') DUM
        READ(10,'(A)') SYSNAME
	write(*,'(1x,a)') SYSNAME
        READ(10,*) MAXSTR
        READ(10,'(A)') UPD_TYPE
        READ(10,*) BSMAGU
        READ(10,*) BSMAGL
        READ(10,*) GSMAGU
        READ(10,*) GSMAGL
        READ(10,*) IASTAR
	print *,'Alignment Star: ',IASTAR
        READ(10,*) NSARC
        READ(10,*) SIGR
        READ(10,*) ITER
        READ(10,*) SIGSCL
        READ(10,*) PSCALE
        READ(10,*) IOPT
        READ(10,*) (PINIT(I),I=1,NMMC)
	print *,'Initial covariances read.'
        DO I=1,NMMC
           MASK(I)=0
        END DO
        READ(10,*,ERR=10,END=10) (PMASK(I),I=1,NMMC)
	print *,'PMASK read.'
C               just in case they've made the mask FLOATING PT.
        DO I=1,NMMC
           MASK(I)=PMASK(I)
        END DO
10      CLOSE(UNIT=10)
        RETURN
        END

C ==========================================================================

        SUBROUTINE MMODEL(MM,NMMC,A1,A2,DELA1,DELA2)
C jlfm/GSFC/920.1...3/19/93...6/18/93
C--------------------------------------------------------------------
C --- MOUNT MODEL CORRECTION SUBROUTINE (for all systems)
C
C       INPUTS:  MM:   MOUNT-MODEL COEFFICIENTS
C                NMMC: NUMBER OF MOUNT MODELLING COEFFICIENTS
C                A1:   CURRENT FIRST ANGLE OF TELESCOPE (DEG)
C                A2:   CURRENT SECOND ANGLE OF TELESCOPE (DEG)
C
C       OUTPUTS: DELA1: MM-CORRECTION FOR FIRST ANGLE (DEG)
C                DELA2: MM-CORRECTION FOR 2ND ANGLE (DEG)
C
C--------------------------------------------------------------------
        IMPLICIT REAL*8 (A-Z)
	  INTEGER MMAX
        PARAMETER (MMAX=30)
        INTEGER I,NMMC
        DIMENSION MM(NMMC),H(2*MMAX)
C
        CALL HMODEL(A1,A2,H)
        DELA1=0.D0
        DELA2=0.D0
        DO I=1,NMMC
           DELA1 = DELA1 + MM(I)*H(2*I-1)
           DELA2 = DELA2 + MM(I)*H(2*I)
        END DO
C
        RETURN
        END

